package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Condition_map.ConditionMapTreeSource

One implementation of LConditionMap, where partitions are hierarchically split in two, based on the order in which they were inserted.

Parameters

module Cond : CONDITION

Signature

Sourcetype 'a t
Sourcemodule Make (L : L) : LConditionMap with module L = L and module Cond = Cond and type t = L.t t