package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Sig.Convert_to_monadicSource

This will help to the transition in a top-down manner, starting from the translation and top-level domain to the lower-level domain.

The idea is to support both interfaces, and use conversion to simplify the support for both. I can have a signature for both domains, and an "AddMonadic" functor to support both domains.

Parameters

module D : BASE

Signature

Sourcemodule Conversion : sig ... end
Sourcemodule Types : sig ... end
Sourcemodule Boolean_Forward : sig ... end
Sourcemodule Binary_Forward : sig ... end
Sourcemodule Enum_Forward : sig ... end