package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Emit_alarmSource

Sourcetype hook = {
  1. hook : 'a. 'a Operator.Alarm.t -> Tracelog.location list -> unit;
}
Sourceval register_alarm_hook : hook -> unit

Register a hook to be called when a new alarm has been emitted. The hook will be called with an alarm and the stack of tracelog locations.

Sourceval emit_alarm : 'a Operator.Alarm.t -> unit

Call all the registered hooks, in order, on the emitted alarm.

Sourceval reset_alarms : unit -> unit

Reset the alarm logs, this should be used only for post analysis