package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Regex.MakeSource

Parameters

module L : LETTER

Signature

Sourcetype letter = L.t
Sourceval empty : t
Sourceval epsilon : t
Sourceval join : t -> t -> t
Sourceval append : t -> letter -> t
Sourceval append_star : t -> t -> t
Sourceval equal : t -> t -> bool
Sourceval hash : t -> int
Sourceval pretty : Format.formatter -> t -> unit
Sourceval pretty_reverse : Format.formatter -> t -> unit

Print a regex in reverse. Useful sometimes when a regex is too big to be * fully printed, and you only want to see the end, you can start printing it * in reverse and interrupt the program.