package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Abstract_interp.CompSource

Signatures for comparison operators ==, !=, <, >, <=, >=.

Sourcetype t =
  1. | Lt
  2. | Gt
  3. | Le
  4. | Ge
  5. | Eq
  6. | Ne
    (*

    comparison operators

    *)
Sourcetype result = truth =
  1. | True
  2. | False
  3. | Unknown
    (*

    result of a comparison

    *)
Sourceval pretty_comp : Format.formatter -> t -> unit
Sourceval inv : t -> t

Inverse relation: a op b <==> ! (a (inv op) b).

Sourceval sym : t -> t

Opposite relation: a op b <==> b (sym op) a.

Sourceval inv_result : result -> result

Given a result r for an operation op, inv_result r is the result that would have been obtained for inv op.