package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module HetHashtblSource

Most of this code is adapated from the OCaml Standard library's implementation of Hashtbl. It is just adapted for polymorphic types. I.E. a 'b t hashtabl binding 'a key to ('a,'b) value (note that 'a is quantified existantially here).

Copyright 1996 Institut National de Recherche en Informatique et en Automatique. distributed under the terms of the GNU Lesser General Public License version 2.1, with the special exception on linking described in the license

Sourcemodule type HETEROGENEOUS_HASHED_TYPE = sig ... end
Sourcemodule type HETEROGENEOUS_SEEDED_HASHED_TYPE = sig ... end
Sourceval is_randomized : unit -> bool
Sourcemodule type S = sig ... end
Sourcemodule MakeSeeded (Key : HETEROGENEOUS_SEEDED_HASHED_TYPE) (Value : PatriciaTree.HETEROGENEOUS_VALUE) : S with type 'key key = 'key Key.t and type ('key, 'a) value = ('key, 'a) Value.t
Sourcemodule Make (Key : HETEROGENEOUS_HASHED_TYPE) (Value : PatriciaTree.HETEROGENEOUS_VALUE) : S with type 'key key = 'key Key.t and type ('key, 'a) value = ('key, 'a) Value.t