package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Bitvector_Of_Integer.MakeSource

Parameters

module I : sig ... end

Signature

include Datatype_sig.S with type t = I.t
Sourcetype t = I.t
Sourceval equal : t -> t -> bool

Any notion of equality is allowed, as long as it is an equivalence relation, and that if a == b, then equal a b.

Sourceval compare : t -> t -> int

compare is a total order, and should be compatible with equal.

Sourceval hash : t -> int

hash requires that equal values have the same hash.

Sourceval bottom : size:Units.In_bits.t -> t
Sourceval is_bottom : size:Units.In_bits.t -> t -> bool
Sourceval top : size:Units.In_bits.t -> t
Sourceval inter : size:Units.In_bits.t -> t -> t -> t
Sourceval join : size:Units.In_bits.t -> t -> t -> t
Sourceval pretty : size:Units.In_bits.t -> Format.formatter -> t -> unit
Sourceval widen : size:Units.In_bits.t -> previous:t -> t -> t
Sourceval includes : size:Units.In_bits.t -> t -> t -> bool
Sourceval includes_or_widen : size:Units.In_bits.t -> previous:t -> t -> bool * t
Sourceval singleton : size:Units.In_bits.t -> Z.t -> t
Sourceval is_singleton : size:Units.In_bits.t -> t -> Z.t option
Sourceval is_empty : size:Units.In_bits.t -> t -> bool

True if the binary cannot be concretized into any value.

Sourceval fold_crop_signed : size:Units.In_bits.t -> t -> inf:Z.t -> sup:Z.t -> 'a -> (Z.t -> 'a -> 'a) -> 'a

Fold on all integers contained in either the signed or unsigned representation of a binary.

Sourceval fold_crop_unsigned : size:Units.In_bits.t -> t -> inf:Z.t -> sup:Z.t -> 'a -> (Z.t -> 'a -> 'a) -> 'a
Sourceval to_known_bits : size:Units.In_bits.t -> t -> Z.t * Z.t
Sourceval to_unsigned_interval : size:Units.In_bits.t -> t -> Z.t * Z.t
Sourceval to_signed_interval : size:Units.In_bits.t -> t -> Z.t * Z.t