package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Autolog.Log_Integer_BackwardSource

Parameters

module F : sig ... end

Signature

Multiply an integer by a constant

This is truncated (C99-like) integer division

Bitwise and, where negative integers are seen as prefixed by infinite ones

Bitwise or, where negative integers are seen as prefixed by infinite ones