package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Conversions.Convert_Block_ForwardSource

Parameters

module C : Conversion
module F : sig ... end

Signature

Sourcetype boolean = F.boolean
Sourcetype offset = F.offset
Sourcetype value = F.value
Sourcetype block = F.block

Size of a block in bytes

Concatenates two blocks

Loads (extracts) a value of a fixed size at a given index from a block

Stores (writes) a fixed size value of a given index in a block

Sourceval binary_to_block : size:Units.In_bits.t -> (value, block) C.To_Arity.ar1

Converts a fixed size value to a block