package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Conversions.Convert_Integer_ForwardSource

Parameters

module C : Conversion
module F : sig ... end

Signature

Sourcetype integer = F.integer
Sourcetype boolean = F.boolean

Multiply an integer by a constant

This is truncated (C99-like) integer division

Bitwise and, where negative integers are seen as prefixed by infinite ones

Bitwise or, where negative integers are seen as prefixed by infinite ones

Integer constant

These can be defined from the others, but it may be more efficient to implement them directly (no need to build temporary values...). They are also often convenient to use directly.