package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Operator.Malloc_idSource

Unique identifier for malloc sites, which eventually includes all allocations in a C program. We also give string as a convenient name for these allocations.

Unique identifier for malloc sites, which eventually includes all allocations in a C program. We also give string as a convenient name for these allocations.

Sourcetype t = private int * string
Sourceval fresh : string -> t
Sourceval hash : t -> int
Sourceval to_string : t -> string
Sourceval to_int : t -> int
Sourceval compare : t -> t -> int
Sourceval equal : t -> t -> bool