package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Sig.Backward_AritySource

Standard arities for backward transfer functions (used to refined the arguments from information on the result values). These take the result value 'r as argument and return a new-improved value for each argument. They return None when no improvement is possible for that argument.

We generally don't include backward function for symbols of arity 0.

Sourcetype 'r ar0 = 'r -> unit
Sourcetype ('a, 'r) ar1 = 'a -> 'r -> 'a option
Sourcetype ('a, 'b, 'r) ar2 = 'a -> 'b -> 'r -> 'a option * 'b option
Sourcetype ('a, 'b, 'c, 'r) ar3 = 'a -> 'b -> 'c -> 'r -> 'a option * 'b option * 'c option