package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Sig.Forward_AritySource

Standard arities for forward transfer functions: given the arguments, return the results. These match the arities of the concrete functions they represent (but with concrete types substituted for their abstract counterparts).

Sourcetype 'r ar0 = 'r
Sourcetype ('a, 'r) ar1 = 'a -> 'r
Sourcetype ('a, 'b, 'r) ar2 = 'a -> 'b -> 'r
Sourcetype ('a, 'b, 'c, 'r) ar3 = 'a -> 'b -> 'c -> 'r