package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Record_timeSource

A very simple benchmarking API.

Sourceval record_time : unit -> unit

By default OCaml uses floats to record dates, which makes me worry about precision. This simple C API does not have this defect. bench_return_time returns the number of microseconds since the last call to bench_record_time.

Sourceval return_time : unit -> int