package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Single_value_abstraction.LogSource

Automatically log the calls to a basis in the trace.

Sourcemodule Sig : sig ... end

Logs all calls to Sub using Tracelog.

Does not logs all calls to Sub using Tracelog; directly use Sub.