package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Single_value_abstraction.StatsSource

Automatically computes stats about the number of calls.

Sourcemodule In_bits = Units.In_bits
Sourcemodule IntMap : sig ... end
Sourcetype func_stats = {
  1. nb_args : int;
    (*

    How many arguments does the function take.

    *)
  2. nb_calls : int;
    (*

    How many times the function was called.

    *)
  3. nb_constant_returns : int;
    (*

    How many times the return value was constant.

    *)
  4. nb_constant_calls : int IntMap.t;
    (*

    For each nb_constant, how many calls with that number of constant args.

    *)
  5. return_size : int IntMap.t option;
    (*

    For each size, how many calls returning that size, if applicable.

    *)
  6. total_time : (int, Stats.compact) Stats.t;
    (*

    Total time spend in this function, in microseconds, across all calls

    *)
}

Aggregated statistics for each transfer function.

Sourceval micro_units : string list
Sourceval micro_units_with_s : string list
Sourceval nano_units : string list
Sourceval print_stats : Format.formatter -> (string * func_stats) list -> unit

Print individual stats for each function in individual order

Sourceval string_ljust : string -> int -> string
Sourceval pp_spread : ?justify:bool -> string list -> string -> Format.formatter -> (int, 'a) Stats.t -> unit
Sourceval print_summary : Format.formatter -> (string * func_stats) list -> unit

Print a summary comparing each functions total runtime

Sourcemodule Make (Sub : sig ... end) : sig ... end