package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Tracelog.MakeSource

Standard functor: verbosity level is controlled by set_verbosity_level.

Parameters

module Category : sig ... end

Signature

Sourceval error : 'a log -> unit

Printing from most to least important. info and debug levels are not printed by default.

Sourceval warning : 'a log -> unit
Sourceval notice : 'a log -> unit
Sourceval info : 'a log -> unit
Sourceval debug : 'a log -> unit
Sourceval fatal : 'a log -> 'b

Classic error messages. Fatal will raise the Fatal exception.

Sourceval not_yet_implemented : 'a log -> unit
Sourceval trace : 'a log -> ?loc:location -> ?bintrace:'c Syntax_tree.Location_identifier.t -> ?pp_ret:(Format.formatter -> 'b -> unit) -> (unit -> 'b) -> 'b

trace log pp_ret f first display log, then executes f and prints its return value using pp_ret.

Sourceval fatal_handler : ('a -> 'b) -> 'a -> 'b

fatal_handler Fatal handler, will properly collect the backtrace for fatal and print it. It also collects other uncaught exception and returns an error message.