package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module TreemapSource

This implement a map from keys that are in a tree relationship (e.g. a sets of paths) to a lattice.

It is so that after n calls to refine ki li, then a map k is mapped to the intersection of all the li for which there was a refine ki li call, and for which ki is a parent of k.

Sourcemodule type Key = sig ... end
Sourcemodule Make_no_empty (Key : Key) : sig ... end

More efficient version, but without empty.

Sourcemodule Make (Key : Key) : sig ... end