package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module type Treemap.KeySource

Sourcetype t
Sourceval nearest_common_ancestor : t -> t -> t

Nearest common ancestor between two elements in the tree

Sourceval is_prefix : t -> t -> bool

is_prefix a b is true if a an ancestor of b

Sourceval equal : t -> t -> bool
Sourceval pretty : Format.formatter -> t -> unit