package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Units.In_bytesSource

Sizes and offsets in bytes.

include Unit
Sourcetype t = private int
Sourceval compare : t -> t -> int
Sourceval of_int : int -> t
Sourceval to_int : t -> int
Sourceval (+) : t -> t -> t
Sourceval (-) : t -> t -> t
Sourceval (<=) : t -> t -> bool
Sourceval (<) : t -> t -> bool
Sourceval (>=) : t -> t -> bool
Sourceval (>) : t -> t -> bool
Sourceval zero : t
Sourceval one : t
Sourceval in_bits : t -> In_bits.t
Sourceval times : int -> t -> t

times k x is k*x. This is useful when multiplying array length with the size of array elements. Don't use this for conversion to sizes in bits.