package codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module While_ast.VarSource

For referring to named storage locations, we use the Var module which ensures that each variable of type t is a record with name and a unique internal identifier to prevent any clashes.

Sourcetype t = {
  1. name : string;
  2. id : int;
}

Here variables with equal names are differentiated using integers through id

Sourceval of_string : string -> t
Sourceval to_int : t -> int
Sourceval pp : Format.formatter -> t -> unit
Sourceval compare : t -> t -> int