Source file condition.ml
1
2
3
4
5
6
7
8
9
10
11
12
13
14
15
16
17
18
19
20
21
22
23
24
25
26
27
28
29
30
31
32
33
34
35
36
37
38
39
40
41
42
43
44
45
46
47
48
49
50
51
52
53
54
55
56
57
58
59
60
61
62
63
64
65
66
67
68
69
70
71
72
73
74
75
76
77
78
79
80
81
82
83
84
85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
108
109
110
111
112
113
114
115
116
117
118
119
120
121
122
123
124
125
126
127
128
129
130
131
132
133
134
135
136
137
138
139
140
141
142
143
144
145
146
147
148
149
150
151
152
153
154
155
156
157
158
159
160
161
162
163
164
165
166
167
168
169
170
171
172
173
174
175
176
177
178
179
180
181
182
183
184
185
186
187
188
189
190
191
192
193
194
195
196
197
198
199
200
201
202
203
204
205
206
207
208
209
210
211
212
213
214
215
216
217
218
219
220
221
222
223
224
225
226
227
228
229
230
231
232
233
234
235
236
237
238
239
240
241
242
243
244
245
246
247
248
249
250
251
252
253
254
255
256
257
258
259
260
261
262
263
264
265
266
267
268
269
270
271
272
273
274
275
276
277
278
279
280
281
282
283
284
285
286
287
288
289
290
291
292
293
294
295
296
297
298
299
300
301
302
303
304
305
306
307
308
309
310
311
312
313
314
315
316
317
318
319
320
321
322
323
324
325
326
327
328
329
330
331
332
333
334
335
336
337
338
339
340
341
342
343
344
345
346
347
348
349
350
351
352
353
354
355
356
357
358
359
360
361
362
363
364
365
366
367
368
369
370
371
372
373
374
375
376
377
378
379
380
381
382
383
384
385
386
387
388
389
390
391
392
393
394
395
396
397
398
399
400
401
402
403
404
405
406
407
408
409
410
411
412
413
414
415
416
417
418
419
420
421
422
423
424
425
426
427
428
429
430
431
432
433
434
435
436
437
438
439
440
441
442
443
444
445
446
447
448
449
450
451
452
453
454
455
456
457
458
459
460
461
462
463
464
465
466
467
468
469
470
471
472
473
474
475
476
477
478
479
480
481
482
483
484
485
486
487
488
489
490
491
492
493
494
495
496
497
498
499
500
501
502
503
504
505
506
507
508
509
510
511
512
513
514
515
516
517
518
519
520
521
522
523
524
525
526
527
528
529
530
531
532
533
534
535
536
537
538
539
540
541
542
543
544
545
546
547
548
549
550
551
552
553
554
555
556
557
558
559
560
561
562
563
564
565
566
567
568
569
570
571
572
573
574
575
576
577
578
579
580
581
582
583
584
585
586
587
588
589
590
591
592
593
594
595
596
597
598
599
600
601
602
603
604
605
606
607
608
609
610
611
612
613
614
615
616
617
618
619
620
621
622
623
624
625
626
627
628
629
630
631
632
633
634
635
636
637
638
639
640
641
642
643
644
645
646
647
648
649
650
651
652
653
654
655
656
657
658
659
660
661
662
663
664
665
666
667
668
669
670
671
672
673
674
675
676
677
678
679
680
681
682
683
684
685
686
687
688
689
690
691
692
693
694
695
696
697
698
699
700
701
702
703
704
705
706
707
708
709
710
711
712
713
714
715
716
717
718
719
720
721
722
723
724
725
726
727
728
729
730
731
732
733
734
735
736
737
738
739
740
741
742
743
744
745
746
747
748
749
750
751
752
753
754
755
756
757
758
759
760
761
762
763
764
765
766
767
768
769
770
771
772
773
774
775
776
777
778
779
780
781
782
783
784
785
786
787
788
789
790
791
792
793
794
795
796
797
798
799
800
801
802
803
804
805
806
807
808
809
810
811
812
813
814
815
816
817
818
819
820
821
822
823
824
825
826
827
828
829
830
831
832
833
834
835
836
837
838
839
840
841
842
843
844
845
846
847
848
849
850
851
852
853
854
855
856
857
858
859
860
861
862
863
864
865
866
867
868
869
870
871
872
873
874
875
876
877
878
879
880
881
882
883
884
885
886
887
888
889
890
891
892
893
894
895
896
897
898
899
900
901
902
903
904
905
906
907
908
909
910
911
912
913
914
915
916
917
918
919
920
921
922
923
924
925
926
927
928
929
930
931
932
933
934
935
936
937
938
939
940
941
942
943
944
945
946
947
948
949
950
951
952
953
954
955
956
957
958
959
960
961
962
963
964
965
966
967
968
969
970
971
972
973
974
975
976
977
978
979
980
981
982
983
984
985
986
987
988
989
990
991
992
993
994
995
996
997
998
999
1000
1001
1002
1003
1004
1005
1006
1007
1008
1009
1010
1011
1012
1013
1014
1015
1016
1017
1018
1019
1020
1021
1022
1023
1024
1025
1026
1027
1028
1029
1030
1031
1032
1033
1034
1035
1036
1037
1038
1039
1040
1041
1042
1043
1044
1045
1046
1047
1048
1049
1050
1051
1052
1053
1054
1055
1056
1057
1058
1059
1060
1061
1062
1063
1064
1065
1066
1067
1068
1069
1070
1071
1072
1073
1074
1075
1076
1077
1078
1079
1080
1081
1082
1083
1084
1085
1086
1087
1088
1089
1090
1091
1092
1093
1094
1095
1096
1097
1098
1099
1100
1101
1102
1103
1104
1105
1106
1107
1108
1109
1110
1111
1112
1113
1114
1115
1116
1117
1118
1119
1120
1121
1122
1123
1124
1125
1126
1127
1128
1129
1130
1131
1132
1133
1134
1135
1136
1137
1138
1139
1140
1141
1142
1143
1144
1145
1146
1147
1148
1149
1150
1151
1152
1153
1154
1155
1156
1157
1158
1159
1160
1161
1162
1163
1164
1165
1166
1167
1168
1169
1170
1171
1172
1173
1174
1175
1176
1177
1178
1179
1180
1181
1182
1183
1184
1185
1186
1187
1188
1189
1190
1191
1192
1193
1194
1195
1196
1197
1198
1199
1200
1201
1202
1203
1204
1205
1206
1207
1208
1209
1210
1211
1212
1213
1214
1215
1216
1217
1218
1219
1220
1221
1222
1223
1224
1225
1226
1227
1228
1229
1230
1231
1232
1233
1234
1235
1236
1237
1238
1239
1240
1241
1242
1243
1244
1245
1246
1247
1248
1249
1250
1251
1252
1253
1254
1255
1256
1257
1258
1259
1260
1261
1262
1263
1264
1265
1266
1267
1268
1269
1270
1271
1272
1273
1274
1275
1276
1277
1278
1279
1280
1281
1282
1283
1284
1285
1286
1287
1288
1289
1290
1291
1292
1293
1294
1295
1296
1297
1298
1299
1300
1301
1302
1303
1304
1305
1306
1307
1308
1309
1310
1311
1312
1313
1314
1315
1316
1317
1318
1319
1320
1321
1322
1323
1324
1325
1326
1327
1328
1329
1330
1331
1332
1333
1334
1335
1336
1337
1338
1339
1340
1341
1342
1343
1344
1345
1346
1347
1348
1349
1350
1351
1352
1353
1354
1355
1356
1357
1358
1359
1360
1361
1362
1363
1364
1365
1366
1367
1368
1369
1370
1371
1372
1373
1374
1375
1376
1377
1378
1379
1380
1381
1382
1383
1384
1385
1386
1387
1388
1389
1390
1391
1392
1393
1394
1395
1396
1397
1398
1399
1400
1401
1402
1403
1404
1405
1406
1407
1408
1409
1410
1411
1412
1413
1414
1415
1416
1417
1418
1419
1420
1421
1422
1423
1424
1425
1426
1427
1428
1429
1430
1431
1432
1433
1434
1435
1436
1437
1438
1439
1440
1441
1442
1443
1444
1445
1446
1447
1448
1449
1450
1451
1452
1453
1454
1455
1456
1457
1458
1459
1460
1461
1462
1463
1464
1465
1466
1467
1468
1469
1470
1471
1472
1473
1474
1475
1476
1477
1478
1479
1480
1481
1482
1483
1484
1485
1486
1487
1488
1489
1490
1491
1492
1493
1494
1495
1496
1497
1498
1499
1500
1501
1502
1503
1504
1505
1506
1507
1508
1509
1510
1511
1512
1513
1514
1515
1516
1517
1518
1519
1520
1521
1522
1523
1524
1525
1526
1527
1528
1529
1530
1531
1532
1533
1534
1535
1536
1537
1538
1539
1540
1541
1542
1543
1544
1545
1546
1547
1548
1549
1550
1551
1552
1553
1554
1555
1556
1557
1558
1559
1560
1561
1562
1563
1564
1565
1566
1567
1568
1569
1570
1571
1572
1573
1574
1575
1576
1577
1578
1579
1580
1581
1582
1583
1584
1585
1586
1587
1588
1589
1590
1591
1592
1593
1594
1595
1596
1597
1598
1599
1600
1601
1602
1603
1604
1605
1606
1607
1608
1609
1610
1611
1612
1613
1614
1615
1616
1617
1618
1619
1620
1621
1622
1623
1624
1625
1626
1627
1628
1629
1630
1631
1632
1633
1634
1635
1636
1637
1638
1639
1640
1641
1642
1643
1644
1645
1646
1647
1648
1649
1650
1651
1652
1653
1654
1655
1656
1657
1658
1659
1660
1661
1662
1663
1664
1665
1666
1667
1668
1669
1670
1671
1672
1673
1674
1675
1676
1677
1678
1679
1680
1681
1682
1683
1684
1685
1686
1687
1688
1689
1690
1691
1692
1693
1694
1695
1696
1697
1698
1699
1700
1701
1702
1703
1704
1705
1706
1707
1708
1709
1710
1711
1712
1713
1714
1715
1716
1717
1718
1719
1720
1721
1722
1723
1724
1725
1726
1727
1728
1729
1730
1731
1732
1733
1734
1735
1736
1737
1738
1739
1740
1741
1742
1743
1744
1745
1746
1747
1748
1749
1750
1751
1752
1753
(** Possible implementations of conditions for the terms. *)
module Bed = Bdd.Make(struct
type t = int
let compare a b = (a - b)
let pretty = Format.pp_print_int
let hash x = x
let equal a b = a == b
end)
module ConditionMy = struct
include Bed.BDD;;
let var_count = ref 0;;
let fresh() = incr var_count; !var_count;;
let all = one
let disjoint a b = ((&&~) a b) == zero
let empty = zero
let is_included a b = (==>~) a b == one
let inter = (&&~)
let union = (||~)
let is_empty x = x == zero
let is_zero x= x == zero
let is_one x = x == one
let complement x = (!~) x
let var () = Bed.BDD.var @@ fresh()
end
(** This uses the structure of the dominator tree as conditions. It
observes the fact that we never perform intersection on arbitrary
conditions, as conditions represent set of paths, we either assume
a new condition or join existing ones.
This works together with skiplist.ml (fast implementation of useful
algorithms like online nearest_common_ancestor) and treemap.ml
(mapping from tree relations to lattices).
It is a good candidate for introducing scopes and getting rid of
Cudd. Still, when I tested these conditions with
Domain_non_relational, it worked for most benchmarks, but the
performance gains were not fantastic compared to cudd (sometimes
less time, some times more; sometimes a lot less memory, but
sometimes more). *)
module ConditionDom = struct
module Literal = struct
type t = int
let var_count = ref 0;;
let fresh() =
var_count := !var_count + 2;
!var_count
;;
let pretty fmt x =
if x land 1 == 1
then Format.fprintf fmt "%d" (x/2)
else Format.fprintf fmt "!%d" (x/2)
let equal = (==)
let complement x =
if x land 1 == 1
then x land (lnot 1)
else x lor 1
let hash x = x
end;;
module Path = struct
type t' =
| Root
| Append of t * Literal.t
| Join of { a: t; b:t; ancestor:t }
and t = {id:int; content:t';depth:int}
;;
let root = {id=0;content=Root;depth=0};;
let rec pretty fmt = function
| x when x == root -> Format.fprintf fmt "root"
| {content=Append(p,l)} -> Format.fprintf fmt "%a::%a" pretty p Literal.pretty l
| {content=Join{a;b;ancestor}} -> Format.fprintf fmt "join(%a,%a,%a)" pretty a pretty b pretty ancestor
| _ -> assert false
;;
let equal a b = a.id == b.id;;
let count = ref 0;;
let tail x = match x.content with
| Root -> assert false
| Append(x,_) -> x
| Join{ancestor} -> ancestor
let rec nearest_common_ancestor a b =
if a.depth == b.depth
then
if a == b
then a
else nearest_common_ancestor (tail a) (tail b)
else if a.depth < b.depth
then nearest_common_ancestor a (tail b)
else nearest_common_ancestor (tail a) b
;;
let is_prefix a b = equal (nearest_common_ancestor a b) a
let inter a b =
let ab = nearest_common_ancestor a b in
if ab == a then b
else if ab == b then a
else Codex_log.fatal "inter %a %a" pretty a pretty b
;;
let disjoint a b =
let ab = nearest_common_ancestor a b in
if ab == a || ab == b
then false
else true
[@@@warning "-8"]
module AppendHash = Weak.Make(struct
type nonrec t = t
let equal ({content=Append(pa,la)}) ({content=Append(pb,lb)}) =
pa == pb && Literal.equal la lb;;
let hash ({content=Append (pa,la)}) = Hashing.hash2 pa.id @@ Literal.hash la
end)
;;
module JoinHash = Weak.Make(struct
type nonrec t = t
let equal ({content=Join{a=a1;b=b1}}) ({content=Join{a=a2;b=b2}}) =
a1 == a2 && b1 == b2
let hash ({content=Join{a;b}}) = Hashing.hash2 a.id b.id
end)
;;
[@@@warning "+8"]
let weakhash_default_size = 2000;;
let tag_ref = ref 1 ;;
let append_table = AppendHash.create weakhash_default_size;;
let join_table = JoinHash.create weakhash_default_size;;
let append p x =
let tentative =
{id = !tag_ref; content = Append(p,x); depth = p.depth + 1} in
let ret = AppendHash.merge append_table tentative in
(if ret == tentative then incr tag_ref);
ret
;;
let join a b =
let ancestor = nearest_common_ancestor a b in
let tentative =
{id = !tag_ref; content = Join{a;b;ancestor}; depth = ancestor.depth + 1} in
let ret = JoinHash.merge join_table tentative in
(if ret == tentative then incr tag_ref);
ret
;;
let union = join
let of_literal x = append root x;;
end
type t =
| False
| Literal of Literal.t
| Path of Path.t
| Complement of t
let hash = function
| False -> 0
| Literal x -> (Path.of_literal x).id
| Path {id} -> id
| Complement _ -> assert false
;;
let pretty fmt x =
match x with
| False -> Format.fprintf fmt "False"
| Literal x -> Format.fprintf fmt "Literal(%a)" Literal.pretty x
| Path x -> Format.fprintf fmt "Path(%a)" Path.pretty x
| Complement _ -> assert false
let equal a b =
match a,b with
| False, False -> true
| Literal x, Literal y -> Literal.equal x y
| Path x, Path y -> Path.equal x y
| Literal x, Path y -> Path.equal y (Path.of_literal x)
| Path x, Literal y -> Path.equal x (Path.of_literal y)
| _ -> false
;;
let _equal a b =
let res = equal a b in
Codex_log.feedback "equal %a %a res %b" pretty a pretty b res;
res
;;
let var() = Literal (Literal.fresh());;
let empty = False;;
let all = Path Path.root;;
let one = all;;
let complement x =
match x with
| Path x when x == Path.root -> False
| Literal x -> Literal (Literal.complement x)
| False -> Path(Path.root)
| Complement x -> x
| x -> Complement x
;;
let (!~) = complement;;
let is_included a b =
match a,b with
| _ when equal a b -> true
| False, _ -> true
| _, False -> false
| _, Path{content=Root} -> true
| Path{content=Root}, _ -> false
| _ -> Codex_log.feedback "is_included %a %a" pretty a pretty b;
assert false;;
let union a b = match a,b with
| False, x | x, False -> x
| Literal a, Literal b when Literal.equal a (Literal.complement b) -> one
| Literal a, Literal b -> Path(Path.union (Path.of_literal a) (Path.of_literal b))
| Path a, Path b -> Path(Path.union a b)
| Path a, Literal b -> Path(Path.union a (Path.of_literal b))
| Literal a, Path b -> Path(Path.union (Path.of_literal a) b)
| _ -> assert false
let (||~) = union
let nearest_common_ancestor a b = match a,b with
| Path {content=Root}, _ -> a
| _, Path {content=Root} -> b
| (Literal a as res), Literal b when Literal.equal a b -> res
| (Literal _a), Literal _b -> all
| Path a, Literal b -> Path(Path.nearest_common_ancestor a (Path.of_literal b))
| Literal a, Path b -> Path(Path.nearest_common_ancestor (Path.of_literal a) b)
| Path a, Path b -> Path(Path.nearest_common_ancestor a b)
| _ -> Codex_log.fatal "nearest_common_ancestor %a %a" pretty a pretty b
let is_prefix a b = match a,b with
| Path {content=Root}, _ -> true
| _, Path {content=Root} -> false
| (Literal a), Literal b when Literal.equal a b -> true
| (Literal a), Literal b -> false
| Path a, Literal b -> Path.is_prefix a (Path.of_literal b)
| Literal a, Path b -> Path.is_prefix (Path.of_literal a) b
| Path a, Path b -> Path.is_prefix a b
| _ -> Codex_log.fatal "is_prefix %a %a" pretty a pretty b
let inter a b =
match a,b with
| False, _ | _, False -> False
| x, y when x == y -> x
| Literal x, Path {content=Append(p,l)} when Literal.equal x l -> b
| Literal x, Path {content=Append(p,l)} when Literal.equal x (Literal.complement l) -> False
| Path {content=Append(p,l)}, Literal x when Literal.equal x l -> a
| Path {content=Append(p,l)}, Literal x when Literal.equal x (Literal.complement l) -> False
| Literal a, Literal b when Literal.equal a (Literal.complement b) -> False
| Literal a, Literal b ->
let b = Path.append Path.root b in
Path(Path.append b a)
| Literal x, Path p -> Path (Path.append p x)
| Path p, Literal x -> Path (Path.append p x)
| Path a, Path b -> Path(Path.inter a b)
| _ ->
Codex_log.fatal "inter %a %a" pretty a pretty b
;;
let (&&~) = inter
let is_empty x = x = False
let disjoint a b = match a,b with
| False, _ | _, False -> true
| (Literal a), Literal b when Literal.equal a b -> false
| (Literal a), Literal b when Literal.equal a (Literal.complement b) -> true
| Path a, Literal b -> Path.disjoint a (Path.of_literal b)
| Literal a, Path b -> Path.disjoint (Path.of_literal a) b
| Path a, Path b -> Path.disjoint a b
| _ -> Codex_log.fatal "disjoint %a %a" pretty a pretty b
;;
module Log = struct
let disjoint a b =
let res = disjoint a b in
Codex_log.feedback "disjoint %a %a res %b" pretty a pretty b res;
res;;
let inter a b =
let res = inter a b in
Codex_log.feedback "inter %a %a res %a" pretty a pretty b pretty res;
res;;
let union a b =
let res = union a b in
Codex_log.feedback "union %a %a res %a" pretty a pretty b pretty res;
res;;
let is_prefix a b =
let res = is_prefix a b in
Codex_log.feedback "is_prefix %a %a res %b" pretty a pretty b res;
res;;
let nearest_common_ancestor a b =
let res = nearest_common_ancestor a b in
Codex_log.feedback "nearest_common_ancestor %a %a res %a" pretty a pretty b pretty res;
res
;;
end
end
(** A dummy Condition, which creates a new int each time. *)
module ConditionInt = struct
type t = int
let count = ref 0 ;;
let unique() = incr count; !count;;
let pretty _ _ = ()
let all = unique ()
let equal a b = a == b
let empty = unique ()
let is_empty x =x == empty
let inter _ _ = unique()
let (&&~) _ _ = unique()
let union _ _ = unique()
let (||~) _ _ = unique()
let disjoint _ _ = assert false
let is_included _ _ = assert false
let complement _ = unique()
let var () = unique()
let hash x = x
end
(** Condition using Cudd binary-decision diagrams. *)
module ConditionCudd = struct
module C_bdd = Cudd.Bdd()
type t = Cudd.bdd
let all = C_bdd.true_ ()
let empty = C_bdd.false_ ()
let inter = C_bdd.and_
let union = C_bdd.or_
let disjoint = C_bdd.is_inter_empty
let is_empty = C_bdd.is_false
let complement = C_bdd.not_
let equal = C_bdd.is_equal
let is_included = C_bdd.is_included_in
let (&&~) = inter
let (||~) = union
let (!~) = complement
let one = all
let zero = empty
let is_zero = C_bdd.is_false
let is_one = C_bdd.is_true
let var () = C_bdd.newvar ()
let pp_print_var fmt v = Format.fprintf fmt "var%d" v
;;
let pretty fmt x = C_bdd.print pp_print_var fmt x
let hash (x:t) = Hashtbl.hash x
end
module MakeConditionCudd(T:sig end):Condition_map.CONDITION = struct
include ConditionCudd
end
module MakeConditionMapMTBDD(Lattice:sig
include Condition_map.L
val equal: t -> t -> bool
val hash: t -> int
val compare: t -> t -> int
end) = struct
module Maybe_Lattice = struct
include Datatype_sig.Option(struct include Lattice let pretty _ = assert false end)
let some x = Some x
let none = None
end
module MTBDD = struct
module Terminal = Maybe_Lattice
include Bed.MTBDD_Make(Maybe_Lattice);;
include With_Set(struct
type t = Lattice.t
let empty = assert false
let singleton x = Maybe_Lattice.the x
let union = assert false
let add a b = assert false
let pretty _ = assert false
end)
end
type value = Lattice.t
module L = Lattice
module Cond = ConditionMy
type t = MTBDD.t
let find = MTBDD.find
let pretty = MTBDD.pretty
let refine mtbdd ~cond ?(notcond=(Cond.complement cond)) value =
let value = Maybe_Lattice.some value in
MTBDD.update mtbdd cond (fun old_lattice ->
match old_lattice,value with
| None, value -> value
| _, None -> assert false
| Some a, Some b -> assert false)
;;
let create x = assert false
let create_partial = MTBDD.terminal None
end
module type SCONDITIONMAP_MTBDD = sig
type value
include Condition_map.LConditionMap with type Cond.t = ConditionMy.t
and type L.t = value
and type t = value option Bed.mtbdd
module MTBDD: Bed.MTBDD
with type Terminal.t = L.t option
end
module ConditionMapMTBDD = struct
include Condition_map.MakePathInsensitive(ConditionMy)(struct
type 'a t = 'a option Bed.mtbdd
end)
let ar1
(type a) (module Ma:SCONDITIONMAP_MTBDD with type value = a)
(type res) (module Mres:SCONDITIONMAP_MTBDD with type value = res)
cond f ma mold =
let f a old = match a,old with
| None, x -> x
| Some a, None -> Some(f a)
| Some a, Some old -> assert false
in
Bed.map2 (module Ma.MTBDD) (module Mres.MTBDD) (module Mres.MTBDD) f ma mold
;;
let ar2
(type a) (module Ma:SCONDITIONMAP_MTBDD with type value = a)
(type b) (module Mb:SCONDITIONMAP_MTBDD with type value = b)
(type res) (module Mres:SCONDITIONMAP_MTBDD with type value = res)
cond f ma mb mold =
let f a b old = match a,b,old with
| None, _, x | _, None, x -> x
| Some a, Some b, None -> Some(f a b)
| Some a, Some b, Some old -> assert false
in
Bed.map3 (module Ma.MTBDD) (module Mb.MTBDD) (module Mres.MTBDD) (module Mres.MTBDD)
f ma mb mold
;;
end
module MakeConditionMapPartitionPI(Condition:Condition_map.CONDITION) = struct
module ConditionMapPartition = Condition_map.ConditionMapPartition(Condition)
module type SCONDITIONMAP = sig
type value
include Condition_map.LConditionMapFold with type Cond.t = Condition.t
and type L.t = value
and type t = value ConditionMapPartition.t
end
module MakeConditionMap(L:Condition_map.L) =
struct
include ConditionMapPartition.Make(L)
type value = L.t
end
module ConditionMap = struct
module PI = Condition_map.MakePathInsensitive(Condition)(struct type 'a t = 'a ConditionMapPartition.t end)
include PI
end
end
module MakeConditionMapTreePI(Condition:Condition_map.CONDITION) = struct
module ConditionMapTree = Condition_map.ConditionMapTree(Condition)
module MakeConditionMap(L:Condition_map.L) =
struct
include ConditionMapTree.Make(L)
type value = L.t
end
module ConditionMap = struct
module PI = Condition_map.MakePathInsensitive(Condition)(struct type 'a t = 'a ConditionMapTree.t end)
include PI
end
end
module CuddMTBDD = struct
type id = int
type var = int
type 'a t =
| TerminalNone
| Terminal of id * 'a
| Ite of id * ConditionCudd.t * 'a t * 'a t
end
module type SCONDITIONMAP_CUDD_MTBDD = sig
type value
include Condition_map.LConditionMapFold with type Cond.t = ConditionCudd.t
and type L.t = value
and type t = value CuddMTBDD.t
end
module MakeConditionMapCuddMTBDD(L:Condition_map.L) = struct
type t = L.t CuddMTBDD.t
let fresh_id =
let count = ref 0 in
fun () ->
incr count;
!count
;;
open CuddMTBDD
let hash = function
| TerminalNone -> 0
| Terminal(id,_) -> id
| Ite(id,_,_,_) -> id
;;
let equal = (==)
let rec pretty pp fmt = function
| TerminalNone -> Format.fprintf fmt "<none>"
| Terminal(id,a) -> Format.fprintf fmt "%d:%a" id pp a
| Ite(id,c,a,b) -> Format.fprintf fmt "%d:ite(@[<hv>%a@,,%a@,,%a@])" id ConditionCudd.pretty c (pretty pp) a (pretty pp) b
[@@@warning "-8"]
module TerminalHash = Weak.Make(struct
type t = L.t CuddMTBDD.t
let equal (Terminal (_,a)) (Terminal (_,b)) = L.equal a b;;
let hash (Terminal (_,a)) = L.hash a
end)
module IfHash = Weak.Make(struct
type t = L.t CuddMTBDD.t
let equal (Ite(_,c1,then1,else1)) (Ite(_,c2,then2,else2)) =
ConditionCudd.equal c1 c2 && then1 == then2 && else1 == else2
let hash (Ite(_,c,then_,else_)) = Hashing.hash3
(ConditionCudd.hash c) (hash then_) (hash else_)
end)
[@@@warning "+8"]
let weakhash_default_size = 2000;;
let tag_ref = ref 1 ;;
let terminal_table = TerminalHash.create weakhash_default_size;;
let if_table = IfHash.create weakhash_default_size;;
let terminal x =
let tentative = Terminal (!tag_ref, x) in
let ret = TerminalHash.merge terminal_table tentative in
(if ret == tentative then (tag_ref := !tag_ref + 1));
ret
;;
let base_mk cond then_ else_ =
let tentative =
if hash then_ < hash else_
then Ite(!tag_ref, cond, then_,else_)
else Ite(!tag_ref, ConditionCudd.complement cond, else_,then_)
in
let ret = IfHash.merge if_table tentative in
(if ret == tentative then (tag_ref := !tag_ref + 1));
ret
;;
let mk cond then_ else_ =
if ConditionCudd.is_one cond then then_
else if ConditionCudd.is_zero cond then else_
else if equal then_ else_ then then_
else
match then_, else_ with
| _, Ite(_,c,a,b) when then_ == a -> base_mk (ConditionCudd.union cond c) a b
| _, Ite(_,c,a,b) when then_ == b -> base_mk (ConditionCudd.inter c (ConditionCudd.complement cond)) a b
| Ite(_,c,a,b), _ when else_ == a -> base_mk (ConditionCudd.inter cond (ConditionCudd.complement c)) b a
| Ite(_,c,a,b), _ when else_ == b -> base_mk (ConditionCudd.inter cond c) a else_
| Ite(_,c1,a1,b1), Ite(_,c2,a2,b2) when a1 == a2 && b1 == b2 ->
base_mk (ConditionCudd.union
(ConditionCudd.inter cond c1)
(ConditionCudd.inter (ConditionCudd.complement cond) c2)) a1 b1
| _ -> base_mk cond then_ else_
;;
let create_partial = TerminalNone
module Cond = ConditionCudd
module L = L
type value = L.t
module T = struct
type t = L.t CuddMTBDD.t
let hash = hash
let equal = (==)
end
module Find_cache = Ephemeron.K2.Make(T)(ConditionCudd);;
let find_cache = Find_cache.create 117;;
module Find_all_cache = Ephemeron.K1.Make(T);;
let find_all_cache = Find_all_cache.create 117;;
let rec find t cond cur =
try Find_cache.find find_cache (t,cond)
with Not_found ->
let res = match Cond.C_bdd.inspect cond with
| Cond.C_bdd.False -> assert false
| Cond.C_bdd.True ->
let find_all x =
try Find_all_cache.find find_all_cache x
with Not_found ->
let res = match x with
| TerminalNone -> assert false
| Terminal(_,v) -> v
| Ite(_,_,t1,t2) -> assert false
in Find_all_cache.replace find_all_cache x res;
res
in find_all t
| Cond.C_bdd.Ifte(var1,then1,else1) -> begin
match t with
| TerminalNone -> assert false
| Terminal(_,a) -> a
| Ite(_,c,then_,else_) -> begin
let curthen = ConditionCudd.inter cur c in
let curelse = ConditionCudd.inter cur (ConditionCudd.complement c) in
match (Cond.C_bdd.is_inter_empty cond curthen, Cond.C_bdd.is_inter_empty cond curelse) with
| false, false -> assert false
| true, false -> find else_ cond curelse
| false, true -> find then_ cond curthen
| true, true -> assert false
end
end
in Find_cache.replace find_cache (t,cond) res;
res
;;
let find t cond = find t cond ConditionCudd.one;;
module Refine_cache = Find_cache;;
module Loop_cache = Ephemeron.K2.Make(ConditionCudd)(ConditionCudd);;
module T2_cache = Ephemeron.K2.Make(T)(T);;
let rec replace_all value = function
| TerminalNone -> terminal value
| Terminal(_,a) -> assert false
| Ite(_,c,then_,else_) ->
mk c (replace_all value then_) (replace_all value else_)
;;
module Value_cache = Ephemeron.K1.Make(L);;
let refine_cache = Value_cache.create 30;;
let refine ~inter (t:L.t CuddMTBDD.t) ~(cond:ConditionCudd.t) ?notcond (value:L.t) =
let refine_cache,loop_cache =
try Value_cache.find refine_cache value
with Not_found ->
let refine_cache = Refine_cache.create 3 in
let loop_cache = T2_cache.create 3 in
(refine_cache,loop_cache)
in
let rec refine t ~cond =
try Refine_cache.find refine_cache (t,cond)
with Not_found ->
let res =
match Cond.C_bdd.inspect cond with
| Cond.C_bdd.True ->
replace_all value t
| Cond.C_bdd.False -> t
| Cond.C_bdd.Ifte(var1,then1,else1) -> begin
match t with
| TerminalNone -> mk cond (terminal value) t
| Terminal(_,a) -> mk cond (terminal @@ inter a value) t
| Ite(_,c2,then2,else2) ->
let c1 = cond in
if ConditionCudd.equal c1 c2 then
mk c2 (replace_all value then2) else2
else if ConditionCudd.equal (ConditionCudd.complement c1) c2 then
mk c2 then2 (replace_all value else2)
else begin
let loop_cache =
try T2_cache.find loop_cache (then2,else2)
with Not_found ->
let res = Loop_cache.create 3 in
T2_cache.replace loop_cache (then2,else2) res;
res
in
let rec loop c1 c2 =
try Loop_cache.find loop_cache (c1,c2)
with Not_found ->
let res =
if ConditionCudd.equal c1 c2
then
mk c2 (replace_all value then2) else2
else if ConditionCudd.equal c1 (ConditionCudd.complement c2)
then
mk c2 then2 (replace_all value else2)
else match Cond.C_bdd.inspect c1 with
| Cond.C_bdd.False -> mk c2 then2 else2
| Cond.C_bdd.True -> mk c2 (replace_all value then2) (replace_all value else2)
| Cond.C_bdd.Ifte(var1,then1,else1) -> begin
match Cond.C_bdd.inspect c2 with
| Cond.C_bdd.True -> refine then2 ~cond:c1
| Cond.C_bdd.False -> refine else2 ~cond:c1
| Cond.C_bdd.Ifte(var2,condthen2,condelse2) ->
if var1 == var2 then
mk (Cond.C_bdd.var var1)
(loop then1 condthen2)
(loop else1 condelse2)
else if var1 < var2 then
mk (Cond.C_bdd.var var1)
(loop then1 c2)
(loop else1 c2)
else
mk (Cond.C_bdd.var var2)
(loop c1 condthen2)
(loop c1 condelse2)
end
in Loop_cache.replace loop_cache (c1,c2) res;
res
in loop c1 c2
end
end
in Refine_cache.replace refine_cache (t,cond) res;
res
in refine t ~cond
;;
let refine t ~cond ?notcond v =
let res = refine t ~cond ?notcond v in
res;;
let fold_with_cond _ = assert false
end
module CuddTree = struct
type id = int
type 'a t =
| TerminalNone
| Terminal of id * 'a
| Ite of id * ConditionCudd.t * 'a t * 'a t
end
module type SCONDITIONMAP_CUDD_TREE = sig
type value
include Condition_map.LConditionMapFold with type Cond.t = ConditionCudd.t
and type L.t = value
and type t = value CuddTree.t
end
module MakeConditionMapCuddTree(L:Condition_map.L) = struct
type t = L.t CuddTree.t
let fresh_id =
let count = ref 0 in
fun () ->
incr count;
!count
;;
open CuddTree
let hash = function
| TerminalNone -> 0
| Terminal(id,_) -> id
| Ite(id,_,_,_) -> id
;;
let equal = (==)
[@@@warning "-8"]
module TerminalHash = Weak.Make(struct
type t = L.t CuddTree.t
let equal (Terminal (_,a)) (Terminal (_,b)) = L.equal a b;;
let hash (Terminal (_,a)) = L.hash a
end)
module IfHash = Weak.Make(struct
type t = L.t CuddTree.t
let equal (Ite(_,c1,then1,else1)) (Ite(_,c2,then2,else2)) =
ConditionCudd.equal c1 c2 && then1 == then2 && else1 == else2
let hash (Ite(_,c,then_,else_)) = Hashing.hash3
(ConditionCudd.hash c) (hash then_) (hash else_)
end)
[@@@warning "+8"]
let weakhash_default_size = 2000;;
let tag_ref = ref 1 ;;
let terminal_table = TerminalHash.create weakhash_default_size;;
let if_table = IfHash.create weakhash_default_size;;
let terminal x =
let tentative = Terminal (!tag_ref, x) in
let ret = TerminalHash.merge terminal_table tentative in
(if ret == tentative then (tag_ref := !tag_ref + 1));
ret
;;
let rec mk var then_ else_ =
assert (var != ConditionCudd.one);
assert (var != ConditionCudd.zero);
if equal then_ else_ then then_
else match then_, else_ with
| _, Ite(_,c,a,b) when then_ == a -> mk (ConditionCudd.union var c) a b
| _, Ite(_,c,a,b) when then_ == b -> mk (ConditionCudd.inter c (ConditionCudd.complement var)) a b
| Ite(_,c,a,b), _ when else_ == a -> mk (ConditionCudd.inter var (ConditionCudd.complement c)) b a
| Ite(_,c,a,b), _ when else_ == b -> mk (ConditionCudd.inter var c) a else_
| Ite(_,c1,a1,b1), Ite(_,c2,a2,b2) when a1 == a2 && b1 == b2 ->
mk (ConditionCudd.union
(ConditionCudd.inter var c1)
(ConditionCudd.inter (ConditionCudd.complement var) c2)) a1 b1
| _ ->
let tentative =
if hash then_ < hash else_
then Ite(!tag_ref, var, then_,else_)
else Ite(!tag_ref, ConditionCudd.complement var, else_,then_)
in
let ret = IfHash.merge if_table tentative in
(if ret == tentative then (tag_ref := !tag_ref + 1));
ret
;;
let rec _pretty pp fmt = function
| TerminalNone -> Format.fprintf fmt "<none>"
| Terminal(id,a) -> Format.fprintf fmt "%d:%a" id pp a
| Ite(id,c,a,b) -> Format.fprintf fmt "%d:ite(@[<v>%a@ ,%a@ ,%a@])" id ConditionCudd.pretty c (_pretty pp) a (_pretty pp) b;;
let pretty fmt _ = ();;
let create_partial = TerminalNone
let rec refine ~inter (t:L.t CuddTree.t) ~cond cur (value:L.t) =
if ConditionCudd.C_bdd.is_inter_empty cond cur
then t
else if ConditionCudd.C_bdd.is_included_in cur cond
then
let rec replace_all = function
| TerminalNone -> terminal value
| Terminal(_,a) -> terminal (inter a value)
| Ite(_,c,then_,else_) ->
mk c (replace_all then_) (replace_all else_)
in replace_all t
else match t with
| TerminalNone -> mk cond (terminal value) TerminalNone
| Terminal(_,a) -> mk cond (terminal (inter a value)) t
| Ite(_,c,then_,else_) ->
let curthen = ConditionCudd.inter cur c in
let curelse = ConditionCudd.inter cur (ConditionCudd.complement c) in
mk c
(refine ~inter then_ ~cond curthen value)
(refine ~inter else_ ~cond curelse value)
;;
let refine t ~cond ?notcond v = refine t ~cond ConditionCudd.one v;;
let refine t ~cond ?notcond v =
let res = refine t ~cond ?notcond v in
res;;
let rec find t cur cond = match t with
| TerminalNone -> assert false
| Terminal(_,v) -> v
| Ite(_,c,then_,else_) ->
let curthen = ConditionCudd.inter cur c in
let curelse = ConditionCudd.inter cur (ConditionCudd.complement c) in
match (ConditionCudd.C_bdd.is_inter_empty cond curthen, ConditionCudd.C_bdd.is_inter_empty cond curelse)
with
| false, false -> assert false
| true, false -> find else_ curelse cond
| false, true -> find then_ curthen cond
| true, true -> assert false
;;
let find t cond =
if ConditionCudd.is_empty cond
then assert false
else find t ConditionCudd.one cond
;;
module Cond = ConditionCudd
module L = L
type value = L.t
let rec fold_with_cond t cond acc f cur =
if Cond.C_bdd.is_inter_empty cur cond then acc
else match t with
| TerminalNone -> assert false
| Terminal(_,v) -> f v (ConditionCudd.inter cur cond) acc
| Ite(_,c,then_,else_) ->
let curthen = ConditionCudd.inter cur c in
let curelse = ConditionCudd.inter cur (ConditionCudd.complement c) in
let acc = fold_with_cond then_ cond acc f curthen in
let acc = fold_with_cond else_ cond acc f curelse in
acc
;;
let fold_with_cond t cond acc f = fold_with_cond t cond acc f ConditionCudd.one;;
end
module CuddPIPartition = struct
module Condition = ConditionCudd
include MakeConditionMapPartitionPI(Condition)
end
module DomPIPartition = struct
module Condition = ConditionDom
module M = Treemap.Make(Condition)
module type SCONDITIONMAP = sig
type value
end
module ConditionMap = struct
type 'a t = 'a M.t
exception Never_refined
let find cond a =
try M.find cond a
with Not_found -> raise Never_refined
let join _ _ = assert false
let ar0 (module Lres:SCONDITIONMAP) ~interres
cond f old =
M.refine cond ~inter:interres ~join f old
;;
let ar1
(module La:SCONDITIONMAP) ~joina ~bottoma
(module Lres:SCONDITIONMAP) ~interres
cond f a old =
let av = find cond a in
M.refine cond ~inter:interres ~join (f av) old
;;
let ar2
(module La:SCONDITIONMAP) ~joina ~bottoma
(module Lb:SCONDITIONMAP) ~joinb ~bottomb
(module Lres:SCONDITIONMAP) ~interres
cond f a b old =
let av = find cond a in
let bv = find cond b in
M.refine cond ~inter:interres ~join (f av bv) old
;;
let nondet_disjoint
(module L:SCONDITIONMAP)
~conda ~notconda ~cma ~condb ~notcondb ~cmb ~join ~bottom ~inter ~old =
let av = find conda cma in
let bv = find condb cmb in
let res = M.refine conda ~inter ~join av old in
let res = M.refine condb ~inter ~join bv res in
res
;;
let nondet_non_disjoint
(module L:SCONDITIONMAP)
~conda ~cma ~condb ~cmb ~condaorb ~notcondaorb ~join ~bottom ~inter ~old =
let av = find conda cma in
let bv = find condb cmb in
let abv = join av bv in
let res = M.refine condaorb ~inter ~join abv old in
res
;;
let changed inter cond r = function
| None -> Condition.empty, r
| Some x -> cond, M.refine cond ~inter ~join x r
;;
let ar1_bwd
(module La:SCONDITIONMAP) ~joina ~bottoma ~intera
(module Lres:SCONDITIONMAP) ~joinres ~bottomres
cond f a res =
let va = find cond a in
let v = find cond res in
let newva = f va v in
changed intera cond a newva
;;
let ar2_bwd
(module La:SCONDITIONMAP) ~joina ~bottoma ~intera
(module Lb:SCONDITIONMAP) ~joinb ~bottomb ~interb
(module Lres:SCONDITIONMAP) ~joinres ~bottomres
cond f a b res =
let va = find cond a in
let vb = find cond b in
let v = find cond res in
let newva, newvb = f va vb v in
(changed intera cond a newva,
changed interb cond b newvb)
;;
end
module MakeConditionMap(L:sig
include Condition_map.L
val pretty: Format.formatter -> t -> unit
end
)= struct
type value = L.t
type t = L.t ConditionMap.t
let create_partial = M.empty
let find ~join ~bottom x key = ConditionMap.find key x;;
let refine ~inter map ~cond value = M.refine cond ~inter ~join:(fun _ _ -> assert false) value map
let pretty = M.make_pretty L.pretty
let find ~join ~bottom x key =
let res = find ~join ~bottom x key in
Codex_log.feedback "Finding %a %a res %a" Condition.pretty key pretty x L.pretty res;
res
;;
let refine ~inter map ~cond value =
let res = refine ~inter map ~cond value in
Codex_log.feedback "Refining %a %a %a res %a" Condition.pretty cond L.pretty value pretty map pretty res;
res
;;
end
end
module CuddPITree = struct
module Condition = ConditionCudd
module ConditionMap = struct
include Condition_map.MakePathInsensitive(ConditionCudd)(struct
type 'a t = 'a CuddTree.t
end)
end
module MakeConditionMap = MakeConditionMapCuddTree
module type SCONDITIONMAP = SCONDITIONMAP_CUDD_TREE
end
module CuddPIMTBDD = struct
module Condition = ConditionCudd
module ConditionMap = struct
include Condition_map.MakePathInsensitive(ConditionCudd)(struct
type 'a t = 'a CuddMTBDD.t
end)
end
module MakeConditionMap = MakeConditionMapCuddMTBDD
module type SCONDITIONMAP = SCONDITIONMAP_CUDD_MTBDD
end
module HomeMadeBDDPartitionPI = struct
module Condition = ConditionMy
include MakeConditionMapPartitionPI(Condition)
end
module HomeMadeMTBDD = struct
module Condition = ConditionMy
module MakeConditionMap = MakeConditionMapMTBDD
module type SCONDITIONMAP = SCONDITIONMAP_MTBDD
module ConditionMap = ConditionMapMTBDD
end